package de.firstdata.ipg.api.client;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.net.ssl.SSLContext;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.MimeHeaders;
import javax.xml.soap.Name;
import javax.xml.soap.SAAJResult;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Consts;
import org.apache.http.HttpEntity;
import org.apache.http.HttpException;
import org.apache.http.HttpHost;
import org.apache.http.HttpResponse;
import org.apache.http.HttpVersion;
import org.apache.http.StatusLine;
import org.apache.http.auth.AUTH;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.Credentials;
import org.apache.http.auth.MalformedChallengeException;
import org.apache.http.auth.NTCredentials;
import org.apache.http.auth.NTUserPrincipal;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.AuthCache;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.config.SocketConfig;
import org.apache.http.conn.ConnectionKeepAliveStrategy;
import org.apache.http.conn.routing.HttpRoute;
import org.apache.http.conn.routing.RouteInfo.LayerType;
import org.apache.http.conn.routing.RouteInfo.TunnelType;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.entity.ContentType;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.auth.NTLMScheme;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.message.BasicHeader;
import org.apache.http.protocol.HttpContext;
import org.apache.http.ssl.SSLContexts;
import org.apache.http.util.EntityUtils;
import org.w3c.dom.NodeList;

import de.firstdata.common.utils.IO;
import de.firstdata.ipg.api.client.action.IPGApiAction;
import de.firstdata.ipg.api.client.action.object.IPGAPIActionObject;
import de.firstdata.ipg.api.client.exception.ClientException;
import de.firstdata.ipg.api.client.exception.InternalException;
import de.firstdata.ipg.api.client.exception.MerchantException;
import de.firstdata.ipg.api.client.exception.ProcessingException;
import de.firstdata.ipg.api.client.exception.ServerException;
import de.firstdata.ipg.api.client.subsequenttransaction.IPGApiSubsequentTransaction;
import de.firstdata.ipg.api.client.subsequenttransaction.object.IPGApiSubsequentTransactionObject;
import de.firstdata.ipg.api.client.transaction.IPGApiTransaction;
import de.firstdata.ipg.api.client.transaction.object.IPGApiMerchantDetailsObject;
import de.firstdata.ipg.api.client.transaction.object.IPGApiTransactionObject;
import de.firstdata.ipg.api.schema.IPGApiActionRequest;
import de.firstdata.ipg.api.schema.IPGApiActionResponse;
import de.firstdata.ipg.api.schema.IPGApiOrderRequest;
import de.firstdata.ipg.api.schema.IPGApiOrderResponse;
import de.firstdata.ipg.api.schema.InquiryCardInformation;
import de.firstdata.ipg.api.schema.InquiryOrder;
import de.firstdata.ipg.api.schema.ObjectFactory;
import de.firstdata.ipg.api.schema.RecurringPayment;
import de.firstdata.ipg.api.schema.RequestCardRateForDCC;
import de.firstdata.ipg.api.schema.RequestMerchantRateForDynamicPricing;
import de.firstdata.ipg.api.schema.StoreHostedData;

/**
 * <code>IPGApiClient</code> represents the core class of the client package.
 * Transaction or Action instances can be passed to the
 * {@link IPGApiClient#commitTransaction(IPGApiTransactionObject)} operation or
 * the {@link IPGApiClient#commitAction(IPGApiActionObject)} operation which
 * builds a SOAP message and sends it to the First Data API Web Service.
 */
public class IPGApiClient {

	private static final Log log = LogFactory.getLog(IPGApiClient.class);

	private static final String[] SUPPORTED_PROTOCOLS = new String[] {"TLSv1", "TLSv1.1", "TLSv1.2"};

	// add NTLM and kerberos later if neccessary
	private static final Collection<String> PREFERRED_AUTHSCHEMES = Collections
			.unmodifiableCollection(Arrays.asList(AuthSchemes.NTLM, AuthSchemes.BASIC, AuthSchemes.DIGEST));

	private static final String XXX;

	private static final Pattern CARD_CODE_VALUE = Pattern.compile("^\\s*(.+(<|:)CardCodeValue>)(\\d+)(</[^:>]*:?CardCodeValue>.+)\\s*$",
			Pattern.DOTALL + Pattern.MULTILINE + Pattern.UNICODE_CASE);

	private static final Pattern CREDIT_CARD_NUMBER = Pattern.compile("^\\s*(.+(<|:)CardNumber>\\d{6})(\\d+)(\\d{4}</[^:>]*:?CardNumber>.+)\\s*$",
			Pattern.DOTALL + Pattern.MULTILINE + Pattern.UNICODE_CASE);

	private static final Pattern TRACK_DATA = Pattern.compile("^\\s*(.+(<|:)TrackData>)(\\d+)(</[^:>]*:?TrackData>.+)\\s*$",
			Pattern.DOTALL + Pattern.MULTILINE + Pattern.UNICODE_CASE);

	private HttpHost proxy;

	protected CloseableHttpClient client;

	private PoolingHttpClientConnectionManager poolingClientConnectionManager;

	private URI ipgApiUri;

	private String ipgApiUrlString;

	private JAXBContext context;

	private String proxyHost;

	private Integer proxyPort;

	private Credentials proxyCredentials;

	private int maxConnectionsPerRoute = 2;

	private String password;

	private int socketTimeoutMs = 60 * 1000;

	private String storeID;

	private byte[] keyStore;

	private String keyStorePW;

	private byte[] trustStore;

	private String trustStorePassword;

	private String userID = "1";

	private String userMidfix = "._.";

	private String userPrefix = "WS";

	private int maxTotalConnections = 20;

	private int connectionPoolTimeout = 15 * 1000;

	private final ConnectionKeepAliveStrategy connectionKeepAliveStrategy = new ConnectionKeepAliveStrategy() {

		@Override
		public long getKeepAliveDuration(final HttpResponse httpResponse, final HttpContext arg1) {
			final int status = httpResponse.getStatusLine().getStatusCode();
			return status != 200 && status != 202 && status != 500 ? 1 : 0;
		}
	};

	static {
		final StringBuilder sb = new StringBuilder();
		for (int i = 0; i < 261; i++) {
			sb.append("X");
		}
		XXX = sb.toString();
	}

	public IPGApiClient() {
		// OK
	}

	/**
	 * creates an instance of the IPGApiClient taking all parameters which are
	 * necessary for communicating with the First Data API Web Service
	 *
	 * @param ipgApiUrl
	 *            the URL of the First Data API Web Service
	 * @param storeId
	 *            your store ID
	 * @param password
	 *            your password
	 * @param keyStore
	 *            a byte array containing your key store
	 * @param keyStorePW
	 *            the password with which your key store can be accessed
	 * @throws IOException
	 * @throws JAXBException
	 * @throws URISyntaxException
	 * @throws GeneralSecurityException
	 */
	public IPGApiClient(final String ipgApiUrl, final String storeId, final String password, final byte[] keyStore, final String keyStorePW)
			throws IOException, JAXBException, URISyntaxException {
		this.ipgApiUrlString = ipgApiUrl;
		this.storeID = storeId;
		this.password = password;
		this.keyStore = keyStore;
		this.keyStorePW = keyStorePW;
		try {
			init();
		} catch (final GeneralSecurityException e) {
			throw new IOException(e);
		}
	}

	public URI getURI() {
		return this.ipgApiUri;
	}

	/**
	 * If you used a setter method, you must call the init method to create a
	 * new connection to the web service. Using the setter methods without the
	 * init method have no effect.
	 *
	 * @throws MalformedURLException
	 * @throws IOException
	 * @throws JAXBException
	 * @throws URISyntaxException
	 * @throws GeneralSecurityException
	 * @throws ClientException
	 */
	public void init() throws IOException, JAXBException, URISyntaxException, GeneralSecurityException {
		final URL url = new URL(this.ipgApiUrlString);
		this.ipgApiUri = url.toURI();
		initClient(url);
		if (this.context == null) {
			this.context = JAXBContext.newInstance(ObjectFactory.class);
		}
	}

	private void reinit() {
		try {
			init();
		} catch (final GeneralSecurityException | URISyntaxException | IOException | JAXBException e) {
			log.warn("cannot create a new client after an exception. Reason: " + e.getMessage(), e);
		}
	}

	private void initClient(final URL url) throws IOException, GeneralSecurityException {
		int port = url.getPort();
		if (port < 0) {
			port = 443;
		}
		this.poolingClientConnectionManager = new PoolingHttpClientConnectionManager(getSocketFactoryRegistry());
		this.poolingClientConnectionManager.setMaxTotal(this.maxTotalConnections);
		this.poolingClientConnectionManager.setDefaultMaxPerRoute(this.maxConnectionsPerRoute);
		final HttpRoute route = new HttpRoute(new HttpHost(url.getHost(), port), null, this.proxy, true,
				this.proxy == null ? TunnelType.PLAIN : TunnelType.TUNNELLED, LayerType.LAYERED);
		this.poolingClientConnectionManager.setMaxPerRoute(route, this.maxConnectionsPerRoute);
		this.poolingClientConnectionManager.setDefaultSocketConfig(SocketConfig.custom() //
				.setSoKeepAlive(true) //
				.setSoTimeout(this.connectionPoolTimeout) //
				.build());
		final BasicCredentialsProvider credentialsProvider = new BasicCredentialsProvider();
		credentialsProvider.setCredentials(new AuthScope(url.getHost(), port), createClientCredentials());
		final HttpClientBuilder httpClientBuilder = HttpClients.custom() //
				.setDefaultCredentialsProvider(credentialsProvider) //
				.setKeepAliveStrategy(this.connectionKeepAliveStrategy) //
				.setConnectionManager(this.poolingClientConnectionManager); //
		initProxy(httpClientBuilder, credentialsProvider);
		this.client = httpClientBuilder.build();
	}

	private Registry<ConnectionSocketFactory> getSocketFactoryRegistry() throws GeneralSecurityException, IOException {
		final SSLContext sslcontext = SSLContexts.custom() //
				.loadTrustMaterial(createTrustStore(), null) //
				.loadKeyMaterial(createKeyStore(), this.keyStorePW.toCharArray()) //
				.build();
		final SSLConnectionSocketFactory sslConnectionSocketFactory = new SSLConnectionSocketFactory(sslcontext, SUPPORTED_PROTOCOLS, null,
				SSLConnectionSocketFactory.getDefaultHostnameVerifier());
		return RegistryBuilder.<ConnectionSocketFactory> create() //
				.register("http", PlainConnectionSocketFactory.getSocketFactory())//
				.register("https", sslConnectionSocketFactory) //
				.build();
	}

	private KeyStore createTrustStore() throws NoSuchAlgorithmException, CertificateException, IOException, KeyStoreException {
		if (this.trustStore == null) {
			this.trustStore = IO.getBytes("/config/truststore.jks");
		}
		if (this.trustStorePassword == null) {
			this.trustStorePassword = IO.getString("/config/truststore.jks.pw.txt").trim();
		}
		return createKeyStore(this.trustStore, this.trustStorePassword);
	}

	private KeyStore createKeyStore() throws NoSuchAlgorithmException, CertificateException, IOException, KeyStoreException {
		return createKeyStore(this.keyStore, this.keyStorePW);
	}

	private static KeyStore createKeyStore(final byte[] keystore, final String keystorePW)
			throws NoSuchAlgorithmException, CertificateException, IOException, KeyStoreException {
		final KeyStore kS = KeyStore.getInstance(KeyStore.getDefaultType());
		kS.load(new ByteArrayInputStream(keystore), keystorePW.toCharArray());
		return kS;
	}

	private void initProxy(final HttpClientBuilder httpClientBuilder, final CredentialsProvider credentialsProvider) {
		if (this.proxyHost == null || this.proxyPort == null) {
			this.proxy = null;
			return;
		}
		this.proxy = new HttpHost(this.proxyHost, this.proxyPort.intValue(), "http");
		httpClientBuilder.setProxy(this.proxy);
		if (this.proxyCredentials != null) {
			credentialsProvider.setCredentials(new AuthScope(this.proxy.getHostName(), this.proxy.getPort()), this.proxyCredentials);
		}
	}

	private Credentials createClientCredentials() {
		final StringBuilder sb = new StringBuilder();
		if (this.userPrefix != null) {
			sb.append(this.userPrefix);
		}
		if (this.storeID != null) {
			sb.append(this.storeID);
		}
		if (this.userMidfix != null) {
			sb.append(this.userMidfix);
		}
		if (this.userID != null) {
			sb.append(this.userID);
		}
		return new UsernamePasswordCredentials(sb.toString(), this.password);
	}

	/**
	 * Sets the proxy. If host or port is null no proxy will be set. If user or
	 * password is null no credentials (authorisation) will be set. If
	 * workstation and domain are not null a windows credentials will be set
	 * otherwise a normal credentials.
	 *
	 * This method has no effect without calling the method init().
	 *
	 * @param host
	 *            name or ip of the proxy machine
	 * @param port
	 *            port of the proxy
	 * @param user
	 *            authorisated user
	 * @param password
	 *            of the authorisated user
	 * @param workstation
	 *            name or ip of the local machine
	 * @param domain
	 *            windows domain
	 */
	public void setProxy(final String host, final Integer port, final String user, final String password, final String workstation, final String domain) {
		if (host == null || port == null) {
			this.proxyHost = null;
			this.proxyPort = null;
			this.proxyCredentials = null;
			return;
		}
		this.proxyHost = host;
		this.proxyPort = port;
		if (user == null || password == null) {
			this.proxyCredentials = null;
			return;
		}
		this.proxyCredentials = workstation == null || domain == null ? new UsernamePasswordCredentials(user, password)
				: new NTCredentials(user, password, workstation, domain);
	}

	/**
	 * commits a transaction on the First Data API by performing a Web Service
	 * call with the transaction data passed
	 *
	 * @param ipgApiTransaction
	 *            the transaction to be committed
	 * @return the transaction result indicating that the transaction has been
	 *         approved by the credit card processor
	 * @throws ClientException
	 *             thrown when finding problems in processing the transaction
	 *             data
	 * @throws MerchantException
	 *             thrown in case the First Data API has registered your store
	 *             as being closed
	 * @throws ProcessingException
	 *             thrown in case the credit card processor has failed to
	 *             process your transaction due to invalid transaction data -
	 *             for details check the exception instance's attributes by
	 *             calling the appropriate <code>getter</code> methods
	 * @throws SOAPException
	 *             thrown when finding problems in building/reading the SOAP
	 *             request/response message
	 * @throws IOException
	 *             thrown when finding internal problems in processing the
	 *             request/response data
	 * @throws ValidationError
	 *             thrown in case your transaction data is not valid according
	 *             to the XML Schema defining the request/response XML
	 *             structures
	 * @throws ServerException
	 *             thrown when finding problems in processing the returned
	 *             server response data
	 * @throws HttpException
	 *             thrown in case the server responds with an HTTP code other
	 *             than <code>200</code> or <code>500</code>
	 * @throws InternalException
	 *             thrown in case of processing your transaction has resulted in
	 *             an internal system error - note that you should never receive
	 *             this exception, however, if so, please contact support
	 * @throws JAXBException
	 * @throws TransformerException
	 */
	public IPGApiResult commitTransaction(final IPGApiTransaction ipgApiTransaction) throws ServerException, ClientException, SOAPException, HttpException,
			MerchantException, ProcessingException, InternalException, IOException, JAXBException {
		return commitTransaction(ipgApiTransaction, null);
	}

	/**
	 * commits a transaction on the First Data API by performing a Web Service
	 * call with the transaction data passed
	 *
	 * @param ipgApiTransaction
	 *            the transaction to be committed
	 * @param ipgApiMerchantDetails
	 *            the merchant information to be committed
	 * @return the transaction result indicating that the transaction has been
	 *         approved by the credit card processor
	 * @throws ClientException
	 *             thrown when finding problems in processing the transaction
	 *             data
	 * @throws MerchantException
	 *             thrown in case the First Data API has registered your store
	 *             as being closed
	 * @throws ProcessingException
	 *             thrown in case the credit card processor has failed to
	 *             process your transaction due to invalid transaction data -
	 *             for details check the exception instance's attributes by
	 *             calling the appropriate <code>getter</code> methods
	 * @throws SOAPException
	 *             thrown when finding problems in building/reading the SOAP
	 *             request/response message
	 * @throws IOException
	 *             thrown when finding internal problems in processing the
	 *             request/response data
	 * @throws ValidationError
	 *             thrown in case your transaction data is not valid according
	 *             to the XML Schema defining the request/response XML
	 *             structures
	 * @throws ServerException
	 *             thrown when finding problems in processing the returned
	 *             server response data
	 * @throws HttpException
	 *             thrown in case the server responds with an HTTP code other
	 *             than <code>200</code> or <code>500</code>
	 * @throws InternalException
	 *             thrown in case of processing your transaction has resulted in
	 *             an internal system error - note that you should never receive
	 *             this exception, however, if so, please contact support
	 * @throws JAXBException
	 * @throws TransformerException
	 */
	public IPGApiResult commitTransaction(final IPGApiTransaction ipgApiTransaction, final IPGApiMerchantDetailsObject ipgApiMerchantDetails)
			throws ServerException, ClientException, SOAPException, HttpException, MerchantException, ProcessingException, InternalException, IOException,
			JAXBException {
		final IPGApiOrderRequest request = new IPGApiOrderRequest();
		request.setTransaction((IPGApiTransactionObject) ipgApiTransaction);
		request.setMerchantDetails(ipgApiMerchantDetails);
		return new IPGApiResult((IPGApiOrderResponse) commit(request));
	}

	public IPGApiResult commitTransaction(final IPGApiSubsequentTransaction ipgApiSubsequentTransaction,
			final IPGApiMerchantDetailsObject ipgApiMerchantDetails) throws ServerException, ClientException, SOAPException, HttpException, MerchantException,
			ProcessingException, InternalException, IOException, JAXBException {
		final IPGApiOrderRequest request = new IPGApiOrderRequest();
		request.setSubsequentTransaction((IPGApiSubsequentTransactionObject) ipgApiSubsequentTransaction);
		request.setMerchantDetails(ipgApiMerchantDetails);
		return new IPGApiResult((IPGApiOrderResponse) commit(request));
	}

	/**
	 * commits an action on the First Data API by performing a Web Service call
	 * with the action data passed
	 *
	 * @param ipgApiAction
	 *            the action to be committed
	 * @return the action result indicating that the action has been approved
	 * @throws ClientException
	 *             thrown when finding problems in processing the action data
	 * @throws MerchantException
	 *             thrown in case the First Data API has registered your store
	 *             as being closed
	 * @throws ProcessingException
	 *             thrown in case the credit card processor has failed to
	 *             process your action due to invalid transaction data - for
	 *             details check the exception instance's attributes by calling
	 *             the appropriate <code>getter</code> methods
	 * @throws SOAPException
	 *             thrown when finding problems in building/reading the SOAP
	 *             request/response message
	 * @throws IOException
	 *             thrown when finding internal problems in processing the
	 *             request/response data
	 * @throws ValidationError
	 *             thrown in case your action data is not valid according to the
	 *             XML Schema defining the request/response XML structures
	 * @throws ServerException
	 *             thrown when finding problems in processing the returned
	 *             server response data
	 * @throws HttpException
	 *             thrown in case the server responds with an HTTP code other
	 *             than <code>200</code> or <code>500</code>
	 * @throws InternalException
	 *             thrown in case of processing your action has resulted in an
	 *             internal system error - note that you should never receive
	 *             this exception, however, if so, please contact support
	 * @throws JAXBException
	 * @throws HttpException
	 * @throws TransformerException
	 */
	public IPGApiActionResponse commitAction(final IPGApiAction ipgApiAction)
			throws MerchantException, ProcessingException, JAXBException, SOAPException, IOException, ServerException, HttpException {
		return commitAction(ipgApiAction, null);
	}

	/**
	 * commits an action on the First Data API by performing a Web Service call
	 * with the action data passed
	 *
	 * @param ipgApiAction
	 *            the action to be committed
	 * @param ipgApiMerchantDetails
	 *            the merchant information to be committed
	 * @return the action result indicating that the action has been approved
	 * @throws ClientException
	 *             thrown when finding problems in processing the action data
	 * @throws MerchantException
	 *             thrown in case the First Data API has registered your store
	 *             as being closed
	 * @throws ProcessingException
	 *             thrown in case the credit card processor has failed to
	 *             process your action due to invalid transaction data - for
	 *             details check the exception instance's attributes by calling
	 *             the appropriate <code>getter</code> methods
	 * @throws SOAPException
	 *             thrown when finding problems in building/reading the SOAP
	 *             request/response message
	 * @throws IOException
	 *             thrown when finding internal problems in processing the
	 *             request/response data
	 * @throws ValidationError
	 *             thrown in case your action data is not valid according to the
	 *             XML Schema defining the request/response XML structures
	 * @throws ServerException
	 *             thrown when finding problems in processing the returned
	 *             server response data
	 * @throws HttpException
	 *             thrown in case the server responds with an HTTP code other
	 *             than <code>200</code> or <code>500</code>
	 * @throws InternalException
	 *             thrown in case of processing your action has resulted in an
	 *             internal system error - note that you should never receive
	 *             this exception, however, if so, please contact support
	 * @throws JAXBException
	 * @throws HttpException
	 * @throws TransformerException
	 */
	public IPGApiActionResponse commitAction(final IPGApiAction ipgApiAction, final IPGApiMerchantDetailsObject merchantDetails)
			throws MerchantException, ProcessingException, JAXBException, SOAPException, IOException, ServerException, HttpException {
		final IPGApiActionRequest request = new IPGApiActionRequest();
		request.setAction((IPGAPIActionObject) ipgApiAction);
		if (merchantDetails == null) {
			// nothing to do
		} else {
			final RecurringPayment recurringPayment = request.getAction().getRecurringPayment();
			final RequestCardRateForDCC requestCardRateForDCC = request.getAction().getRequestCardRateForDCC();
			final RequestMerchantRateForDynamicPricing requestMerchantRateForDynamicPricing = request.getAction().getRequestMerchantRateForDynamicPricing();
			final StoreHostedData requestStoreHostedData = request.getAction().getStoreHostedData();
			final InquiryOrder requestInquiryOrder = request.getAction().getInquiryOrder();
			final InquiryCardInformation requestInquiryCardInformation = request.getAction().getInquiryCardInformation();
			if (recurringPayment != null) {
				recurringPayment.setMerchantDetails(merchantDetails);
			} else if (requestCardRateForDCC != null) {
				requestCardRateForDCC.setMerchantDetails(merchantDetails);
			} else if (requestMerchantRateForDynamicPricing != null) {
				requestMerchantRateForDynamicPricing.setMerchantDetails(merchantDetails);
			} else if (requestStoreHostedData != null) {
				requestStoreHostedData.setMerchantDetails(merchantDetails);
			} else if (requestInquiryOrder != null) {
				requestInquiryOrder.setMerchantDetails(merchantDetails);
			} else if (requestInquiryCardInformation != null) {
				requestInquiryCardInformation.setMerchantDetails(merchantDetails);
			} else {
				// if merchantDetails not null and none of the action type above
				throw new MerchantException("merchant details not supported for the given action type");
			}
		}
		return (IPGApiActionResponse) commit(request);
	}

	private static String mask(String xml) {
		xml = mask(xml, CARD_CODE_VALUE);
		xml = mask(xml, CREDIT_CARD_NUMBER);
		xml = mask(xml, TRACK_DATA);
		return xml;
	}

	private static String mask(final String request, final Pattern pattern) {
		final Matcher matcher = pattern.matcher(request);
		if (matcher.matches()) {
			return matcher.group(1) + XXX.substring(0, matcher.group(3).length()) + matcher.group(4);
		}
		return request;
	}

	protected Object commit(final Object request)
			throws JAXBException, SOAPException, IOException, MerchantException, ProcessingException, ServerException, HttpException {
		final SOAPMessage requestMessage = MessageFactory.newInstance().createMessage();
		this.context.createMarshaller().marshal(request, new SAAJResult(requestMessage.getSOAPPart().getEnvelope().getBody()));
		try (final ByteArrayOutputStream requestOutStream = new ByteArrayOutputStream()) {
			requestMessage.writeTo(requestOutStream);
			requestOutStream.flush();
			final byte[] bytes = requestOutStream.toByteArray();
			return commit(bytes);
		}
	}

	protected Object commit(final byte[] requestBytes)
			throws IOException, SOAPException, JAXBException, ServerException, MerchantException, ProcessingException, HttpException {
		final HttpPost httpPost = new HttpPost(this.ipgApiUri);
		httpPost.setProtocolVersion(HttpVersion.HTTP_1_1);
		httpPost.setHeader("Content-Type", ContentType.TEXT_XML.withCharset(Consts.UTF_8).toString());
		httpPost.setHeader("CERT_REMOTE_USER", this.createClientCredentials().getUserPrincipal().getName());
		httpPost.setEntity(new ByteArrayEntity(requestBytes));
		if (log.isInfoEnabled()) {
			log.info("request to " + this.ipgApiUri + System.lineSeparator() + XMLTransformer.transform(mask(new String(requestBytes, "UTF-8"))));
		}
		final HttpClientContext httpClientContext = HttpClientContext.create();
		httpClientContext.setRequestConfig(getRequestConfig());
		setAuthCache(httpClientContext, this.ipgApiUri);

		try (final CloseableHttpResponse httpResponse = this.client.execute(httpPost, httpClientContext)) {
			final StatusLine statusLine = httpResponse.getStatusLine();
			if (log.isInfoEnabled()) {
				log.info("response status line: " + statusLine);
			}
			final int status = statusLine.getStatusCode();
			if (status != 200 && status != 202 && status != 500) {
				reinit();
				throw new HttpException("HTTP server has responded with the following error code: " + status);
			}
			final HttpEntity entity = httpResponse.getEntity();
			final byte[] response = EntityUtils.toByteArray(entity);
			if (log.isInfoEnabled()) {
				log.info("response:" + System.lineSeparator() + XMLTransformer.transform(new String(response, "UTF-8")));
			}
			final SOAPBody responseBody = MessageFactory.newInstance().createMessage(new MimeHeaders(), new ByteArrayInputStream(response)).getSOAPBody();
			final Unmarshaller unmarshaller = this.context.createUnmarshaller();
			if (!responseBody.hasFault()) {
				return unmarshaller.unmarshal(new DOMSource(responseBody.getFirstChild()));
			}
			final SOAPFault fault = responseBody.getFault();
			final Name faultCode = fault.getFaultCodeAsName();
			final String faultString = fault.getFaultString();
			if (faultCode == null) {
				throw new ServerException("soap fault code missing - please contact support. Error message: " + faultString);
			}
			if ("Server".equals(faultCode.getLocalName())) {
				throw new ServerException("Server error - please contact support. Error message: " + faultString);
			}
			final int endTypePos = faultString.indexOf(':');
			final String exceptionType;
			if (endTypePos >= 0) {
				exceptionType = faultString.substring(0, endTypePos);
			} else {
				exceptionType = faultString;
			}
			if ("ProcessingException".equals(exceptionType)) {
				final IPGApiOrderResponse ipgResponse = (IPGApiOrderResponse) unmarshaller.unmarshal(new DOMSource(fault.getDetail().getFirstChild()));
				throw new ProcessingException(ipgResponse.getErrorMessage(), ipgResponse);
			}
			if ("MerchantException".equals(exceptionType)) {
				final NodeList nodeList = fault.getDetail().getChildNodes();
				final StringBuilder sb = new StringBuilder();
				for (int i = 0; i < nodeList.getLength(); i++) {
					sb.append(nodeList.item(i).getTextContent());
					sb.append("\n");
				}
				throw new MerchantException(sb.toString().trim());
			}
			throw new InternalException(faultString);
		} catch (final IOException e) {
			reinit();
			throw e;
		}
	}

	private void setAuthCache(final HttpClientContext localContext, final URI uri) {
		// Create AuthCache instance
		final AuthCache authCache = new BasicAuthCache();
		// Generate BASIC scheme object and add it to the local
		// auth cache
		authCache.put(new HttpHost(uri.getHost(), uri.getPort(), uri.getScheme()), new BasicScheme());
		if (this.proxyHost != null) {
			if (this.proxyCredentials instanceof NTUserPrincipal) {
				authCache.put(this.proxy, new NTLMScheme());
			}
			if (this.proxyCredentials instanceof UsernamePasswordCredentials) {
				final BasicScheme proxyAuth = new BasicScheme();
				try {
					proxyAuth.processChallenge(new BasicHeader(AUTH.PROXY_AUTH, "BASIC realm=default"));
				} catch (final MalformedChallengeException e) {
					throw new IllegalArgumentException(e);
				}
				authCache.put(this.proxy, proxyAuth);
			}
		}
		localContext.setAuthCache(authCache);
	}

	private RequestConfig getRequestConfig() {
		return RequestConfig.custom() //
				.setConnectTimeout(this.socketTimeoutMs) //
				.setConnectionRequestTimeout(this.socketTimeoutMs) //
				.setSocketTimeout(this.socketTimeoutMs) //
				.setExpectContinueEnabled(false) //
				.setProxyPreferredAuthSchemes(PREFERRED_AUTHSCHEMES) //
				.setTargetPreferredAuthSchemes(PREFERRED_AUTHSCHEMES) //
				.build();
	}

	/**
	 * This method has no effect without calling the method init().
	 */
	public void setIpgApiUrl(final String value) {
		this.ipgApiUrlString = value;
	}

	public String getIpgApiUrl() {
		return this.ipgApiUrlString;
	}

	/**
	 * This method has no effect without calling the method init().
	 */
	public void setKeyStore(final byte[] value) {
		this.keyStore = value;
	}

	public byte[] getKeyStore() {
		return this.keyStore;
	}

	/**
	 * This method has no effect without calling the method init().
	 */
	public void setKeyStorePassword(final String value) {
		this.keyStorePW = value;
	}

	public String getKeyStorePassword() {
		return this.keyStorePW;
	}

	/**
	 * This should be set to the numbers of terminal You have. We set this to 2.
	 *
	 * For detail information see the apache http client 4.0.3 doc.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setMaxConnectionsPerRoute(final int value) {
		this.maxConnectionsPerRoute = value;
	}

	public int getMaxConnectionsPerRoute() {
		return this.maxConnectionsPerRoute;
	}

	/**
	 * The API store password.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setPassword(final String value) {
		this.password = value;
	}

	public String getPassword() {
		return this.password;
	}

	/**
	 * This method sets the CoreConnectionPNames.SO_TIMEOUT. The apache http
	 * client waits this time after the message is sent to IPG. We set this to
	 * 60 seconds. You should not set this less than 60 seconds.
	 *
	 * For detail information see the apache http client 4.0.3 doc.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setSocketTimeout(final int value) {
		this.socketTimeoutMs = value;
	}

	public int getSocketTimeout() {
		return this.socketTimeoutMs;
	}

	/**
	 * StoreID or technical user. If you use the technical user, than you must
	 * put the storeID into the transactions.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setStoreID(final String value) {
		this.storeID = value;
	}

	public String getStoreID() {
		return this.storeID;
	}

	/**
	 * This method has no effect without calling the method init().
	 */
	public void setTrustStore(final byte[] value) {
		this.trustStore = value;
	}

	public byte[] getTrustStore() {
		return this.trustStore;
	}

	/**
	 * This method has no effect without calling the method init().
	 */
	public void setTrustStorePassword(final String value) {
		this.trustStorePassword = value;
	}

	public String getTrustStorePassword() {
		return this.trustStorePassword;
	}

	/**
	 * We set this to 1
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setUserID(final String value) {
		this.userID = value;
	}

	public String getUserID() {
		return this.userID;
	}

	/**
	 * We set this to ._.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setUserMidfix(final String value) {
		this.userMidfix = value;
	}

	public String getUserMidfix() {
		return this.userMidfix;
	}

	/**
	 * We set this to WS.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setUserPrefix(final String value) {
		this.userPrefix = value;
	}

	public String getUserPrefix() {
		return this.userPrefix;
	}

	/**
	 * For detail information see the apache http client 4.0.3 doc.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setMaxTotalConnections(final int value) {
		this.maxTotalConnections = value;
	}

	public int getMaxTotalConnections() {
		return this.maxTotalConnections;
	}

	/**
	 * This method sets the ConnManagerPNames.TIMEOUT. The apache http client
	 * waits this time for a connection in the connection pool. We set this to
	 * 15 seconds.
	 *
	 * For detail information see the apache http client 4.0.3 doc.
	 *
	 * This method has no effect without calling the method init().
	 */
	public void setConnectionPoolTimeout(final int value) {
		this.connectionPoolTimeout = value;
	}

	public int getTimeout() {
		return this.connectionPoolTimeout;
	}
}
