package de.firstdata.ipg.api.client.action.recurringpayment.object;

import java.util.Date;

import javax.xml.datatype.XMLGregorianCalendar;

import de.firstdata.ipg.api.client.action.object.IPGAPIActionObject;
import de.firstdata.ipg.api.client.transaction.Amount;
import de.firstdata.ipg.api.client.transaction.creditcard.CreditCard;
import de.firstdata.ipg.api.client.transaction.dedirectdebit.DEDirectDebit;
import de.firstdata.ipg.api.schema.Basket;
import de.firstdata.ipg.api.schema.Billing;
import de.firstdata.ipg.api.schema.CardFunctionType;
import de.firstdata.ipg.api.schema.CreditCardData;
import de.firstdata.ipg.api.schema.DEDirectDebitData;
import de.firstdata.ipg.api.schema.Function;
import de.firstdata.ipg.api.schema.GenderType;
import de.firstdata.ipg.api.schema.Payment;
import de.firstdata.ipg.api.schema.RecurringPayment;
import de.firstdata.ipg.api.schema.RecurringPaymentInformation;
import de.firstdata.ipg.api.schema.Shipping;
import de.firstdata.ipg.api.schema.Wallet;
import de.firstdata.ipg.api.schema.WalletType;

abstract class RecurringPaymentObject extends IPGAPIActionObject {

	protected RecurringPaymentObject(final Function function) {
		getRecurringPayment().setFunction(function);
	}

	public void setAmount(final Amount value) {
		setPayment(value);
	}

	public void setBasket(final Basket value) {
		getRecurringPayment().setBasket(value);
	}

	public void setComments(final String value) {
		getRecurringPayment().setComments(value);
	}

	public void setInstallmentCount(final int value) {
		getRecurringPaymentInformation().setInstallmentCount(Integer.valueOf(value));
	}

	public void setInstallmentFrequency(final int value) {
		getRecurringPaymentInformation().setInstallmentFrequency(Integer.valueOf(value));
	}

	public void setInstallmentPeriod(final String value) {
		getRecurringPaymentInformation().setInstallmentPeriod(value);
	}

	public void setMaximumFailures(final int value) {
		getRecurringPaymentInformation().setMaximumFailures(Integer.valueOf(value));
	}

	public void setRecurringStartDate(final String value) {
		getRecurringPaymentInformation().setRecurringStartDate(value);
	}

	public void setRecurringStartDate(final Date value) {
		setRecurringStartDate(getFormattedDate(value));
	}

	public void setWalletType(final WalletType value) {
		getWallet().setWalletType(value);
	}

	public void setWalletID(final String value) {
		getWallet().setWalletID(value);
	}

	@Override
	public void setStoreId(final String storeId) {
		getRecurringPayment().setStoreId(storeId);
	}

	public void setOrderId(final String orderId) {
		getRecurringPayment().setOrderId(orderId);
	}

	protected void setPayment(final Amount amount) {
		if (amount == null) {
			return;
		}
		final Payment payment = getPayment();
		payment.setChargeTotal(amount.total);
		payment.setCurrency(amount.currency);
	}

	public void setCreditCard(final CreditCard creditCard) {
		final CreditCardData data = new CreditCardData();
		data.setCardNumber(creditCard.cardNumber);
		data.setExpMonth(creditCard.expMonth);
		data.setExpYear(creditCard.expYear);
		data.setCardCodeValue(creditCard.cardCodeValue);
		data.setTrackData(creditCard.trackData);
		getRecurringPayment().setCreditCardData(data);
	}

	public void setDEDirectDebit(final DEDirectDebit deDirectDebit) {
		final DEDirectDebitData data = new DEDirectDebitData();
		if (deDirectDebit.iban == null) {
			data.setBankCode(deDirectDebit.bankCode);
			data.setAccountNumber(deDirectDebit.accountNumber);
		} else {
			data.setBIC(deDirectDebit.bic);
			data.setIBAN(deDirectDebit.iban);
		}
		data.setTrackData(deDirectDebit.trackData);
		getRecurringPayment().setDEDirectDebitData(data);
	}

	public void setHostedDataID(final String value) {
		getPayment().getHostedDataID().add(value);
	}

	public void setHostedDataStoreID(final String value) {
		getPayment().setHostedDataStoreID(value);
	}

	public void setReferencedOrderID(final String value) {
		getRecurringPayment().setReferencedOrderId(value);
	}

	public void setCustomerID(final String customerID) {
		getBilling().setCustomerID(customerID);
	}

	public void setName(final String name) {
		getBilling().setName(name);
	}

	public void setCompany(final String company) {
		getBilling().setCompany(company);
	}

	public void setAddress1(final String address1) {
		getBilling().setAddress1(address1);
	}

	public void setAddress2(final String address2) {
		getBilling().setAddress2(address2);
	}

	public void setCity(final String city) {
		getBilling().setCity(city);
	}

	public void setState(final String state) {
		getBilling().setState(state);
	}

	public void setZip(final String zip) {
		getBilling().setZip(zip);
	}

	public void setCountry(final String country) {
		getBilling().setCountry(country);
	}

	public void setPhone(final String phone) {
		getBilling().setPhone(phone);
	}

	public void setFax(final String fax) {
		getBilling().setFax(fax);
	}

	public void setEmail(final String email) {
		getBilling().setEmail(email);
	}

	public void setPersonalNumber(final String personalNumber) {
		getBilling().setPersonalNumber(personalNumber);
	}

	public void setBirthDate(final XMLGregorianCalendar birthDate) {
		getBilling().setBirthDate(birthDate);
	}

	public void setGender(final GenderType gender) {
		getBilling().setGender(gender);
	}

	public void setFirstname(final String firstname) {
		getBilling().setFirstname(firstname);
	}

	public void setSurname(final String surname) {
		getBilling().setSurname(surname);
	}

	public void setStreetName(final String streetName) {
		getBilling().setStreetName(streetName);
	}

	public void setHouseNumber(final String houseNumber) {
		getBilling().setHouseNumber(houseNumber);
	}

	public void setHouseExtension(final String houseExtension) {
		getBilling().setHouseExtension(houseExtension);
	}

	public void setMobilePhone(final String mobilePhone) {
		getBilling().setMobilePhone(mobilePhone);
	}

	@Deprecated
	public void setAddressNumber(final String addressNumber) {
		getBilling().setAddrnum(addressNumber);
	}

	@Deprecated
	public void setType(final String type) {
		getShipping().setType(type);
	}

	public void setShippingAddressName(final String name) {
		getShipping().setName(name);
	}

	public void setShippingAddress1(final String address1) {
		getShipping().setAddress1(address1);
	}

	public void setShippingAddress2(final String address2) {
		getShipping().setAddress2(address2);
	}

	public void setShippingAddressCity(final String city) {
		getShipping().setCity(city);
	}

	public void setShippingAddressState(final String state) {
		getShipping().setState(state);
	}

	public void setShippingAddressZip(final String zip) {
		getShipping().setZip(zip);
	}

	public void setShippingAddressCountry(final String country) {
		getShipping().setCountry(country);
	}

	protected Billing getBilling() {
		Billing billing = getRecurringPayment().getBilling();
		if (billing == null) {
			billing = new Billing();
			getRecurringPayment().setBilling(billing);
		}
		return billing;
	}

	protected Shipping getShipping() {
		Shipping shipping = getRecurringPayment().getShipping();
		if (shipping == null) {
			shipping = new Shipping();
			getRecurringPayment().setShipping(shipping);
		}
		return shipping;
	}

	protected Payment getPayment() {
		Payment payment = getRecurringPayment().getPayment();
		if (payment == null) {
			payment = new Payment();
			getRecurringPayment().setPayment(payment);
		}
		return payment;
	}

	@Override
	public RecurringPayment getRecurringPayment() {
		if (this.recurringPayment == null) {
			this.recurringPayment = new RecurringPayment();
		}
		return this.recurringPayment;
	}

	public Wallet getWallet() {
		Wallet wallet = getRecurringPayment().getWallet();
		if (wallet == null) {
			wallet = new Wallet();
			getRecurringPayment().setWallet(wallet);
		}
		return wallet;
	}

	protected RecurringPaymentInformation getRecurringPaymentInformation() {
		RecurringPaymentInformation recurringPaymentInformation = getRecurringPayment().getRecurringPaymentInformation();
		if (recurringPaymentInformation == null) {
			recurringPaymentInformation = new RecurringPaymentInformation();
			getRecurringPayment().setRecurringPaymentInformation(recurringPaymentInformation);
		}
		return recurringPaymentInformation;
	}

	public void setCardFunction(final CardFunctionType value) {
		getRecurringPayment().setCardFunction(value);
	}
}
