package de.firstdata.ipg.api.client.transaction.creditcard.object;

import de.firstdata.ipg.api.client.IPGApiClient;
import de.firstdata.ipg.api.client.transaction.Amount;
import de.firstdata.ipg.api.client.transaction.creditcard.CCForceTicketTransaction;
import de.firstdata.ipg.api.client.transaction.creditcard.CreditCard;
import de.firstdata.ipg.api.schema.Secure3DAuthenticationRequest;
import de.firstdata.ipg.api.schema.Secure3DVerificationRequest;

/**
 * This class represents a <i>ForceTicket</i> transaction and allows for setting
 * its optional attributes by calling the appropriate <code>setter</code>
 * methods. An instance of this class is created when calling
 * {@link IPGApiClient#createForceTicketTransaction(double, String)}.
 * 
 * @author Andreas Schmid
 */
public final class CCForceTicketTransactionObject extends CCTransactionObject implements CCForceTicketTransaction {

	/**
	 * creates a <i>ForceTicket</i> transaction with the minimum set of
	 * parameters - note that you can add optional fields by calling the
	 * appropriate <code>setter</code> methods on the returned
	 * {@link CCForceTicketTransactionObject} instance
	 * 
	 * @param referenceNumber
	 *            the reference number you have obtained when having authorized
	 *            this transaction e.g. by phone
	 */
	public CCForceTicketTransactionObject(final Amount amount, final CreditCard creditCard, final String referenceNumber) {
		super("forceTicket", amount, creditCard, null, null, null);
		getTransactionDetails().setReferenceNumber(referenceNumber);
	}

	public CCForceTicketTransactionObject(final Amount amount, final String hostedDataID, final String hostedDataStoreID, final String referenceNumber) {
		super("forceTicket", amount, null, hostedDataID, hostedDataStoreID, null);
		getTransactionDetails().setReferenceNumber(referenceNumber);
	}

	@Override
	public void setSecure3DVerificationRequest(final Secure3DVerificationRequest secure3dVerificationRequest) {
		getSecure3DRequest().setSecure3DVerificationRequest(secure3dVerificationRequest);
	}

	@Override
	public void setSecure3DAuthenticationRequest(final Secure3DAuthenticationRequest secure3dAuthenticationRequest) {
		getSecure3DRequest().setSecure3DAuthenticationRequest(secure3dAuthenticationRequest);
	}
}
